// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ILUPreconditioner.h */

#ifndef AMDIS_ILU_PRECONDITIONER_H
#define AMDIS_ILU_PRECONDITIONER_H

#include <vector>
#include "Preconditioner.h"
#include "MemoryManager.h"
#include "CreatorInterface.h"
#include "DOFMatrix.h"

namespace AMDiS {

  template<typename T> class DOFVector;
  template<typename T> class OEMSolver;

  // ============================================================================
  // ===== class ILUPreconditioner ==============================================
  // ============================================================================

  /**
   * \ingroup Solver
   * 
   * \brief
   * ILU (Incomplete LU factorization) preconditioner.
   *
   * The preconditioner is implemented following the book "Iterative Methods for
   * Sparce Linear Systems", second edition, Yousef Saad. The preconditioner is
   * described in chapter 10.3 (algorithm 10.4).
   */
  class ILUPreconditioner : public PreconditionerScal
  {
  public:
    MEMORY_MANAGED(ILUPreconditioner);

    /** \brief
     * Creater class used in PreconditionerMap.
     */     
    class Creator : public PreconditionerScalCreator
    {
    public:
      MEMORY_MANAGED(Creator);

      /** \brief
       * Creates an ILUPreconditioner.
       */
      PreconditionerScal *create() {
	return NEW ILUPreconditioner(size, row);
      };
    };

    /** \brief
     * Constructor.
     */
    ILUPreconditioner(int size = 1, int row = 0)
      : PreconditionerScal(size, row)
    {};

    /** \brief
     * Destructor.
     */
    virtual ~ILUPreconditioner() {};

    /** \brief
     * Realisation of Preconditioner::init
     *
     * In the initalization phase the incomplete LU factorization
     * of the coefficent matrix is calculated.
     */
    void init();

    /** \brief
     * Realisation of Preconditioner::exit
     */
    void exit();
    
    /** \brief
     * Realisation of Preconditioner::precon
     */
    void precon(DOFVector<double>* x);

  private:
    /** \brief
     * This matrix stores the incomplete LU factorization
     * of the coefficient matrix.
     */
    //    DOFMatrix *ilu_;
    Vector<DOFMatrix*> ilu_;

    /** \brief
     * Here we store the diagonal elements of ilu_ matrix. This
     * allows faster computations.
     */
    std::vector<double> diag_;
  };
















  class ILUPreconditionerStd : public PreconditionerScalStd
  {
  public:
    MEMORY_MANAGED(ILUPreconditionerStd);


    /** \brief
     * Constructor.
     */
    ILUPreconditionerStd() {};

    /** \brief
     * Destructor.
     */
    virtual ~ILUPreconditionerStd() {};

    /** \brief
     * Realisation of Preconditioner::init
     *
     * In the initalization phase the incomplete LU factorization
     * of the coefficent matrix is calculated.
     */
    void init();

    /** \brief
     * Realisation of Preconditioner::exit
     */
    void exit();
    
    /** \brief
     * Realisation of Preconditioner::precon
     */
    void precon(std::vector<double>* x);

  private:
    /** \brief
     * This matrix stores the incomplete LU factorization
     * of the coefficient matrix.
     */
    std::vector< std::vector<MatEntry> > ilu_;

    /** \brief
     * Here we store the diagonal elements of ilu_ matrix. This
     * allows faster computations.
     */
    std::vector<double> diag_;
  };

}

#endif // AMDIS_ILU_PRECONDITIONER_H
